var jotai = (function (exports, react, useContextSelector) {
  'use strict';

  var isSSR = typeof window === 'undefined' || /ServerSideRendering/.test(window.navigator && window.navigator.userAgent);
  var useIsoLayoutEffect = isSSR ? function (fn) {
    return fn();
  } : react.useLayoutEffect;

  var appendMap = function appendMap(dst, src) {
    src.forEach(function (v, k) {
      dst.set(k, v);
    });
    return dst;
  }; // create new map from two maps


  var concatMap = function concatMap(src1, src2) {
    return appendMap(new Map(src1), src2);
  };

  var warningObject = new Proxy({}, {
    get: function get() {
      throw new Error('Please use <Provider>');
    },
    apply: function apply() {
      throw new Error('Please use <Provider>');
    }
  });

  var addDependent = function addDependent(dependentsMap, atom, dependent) {
    var dependents = dependentsMap.get(atom);

    if (!dependents) {
      dependents = new Set();
      dependentsMap.set(atom, dependents);
    }

    dependents.add(dependent);
  };

  var deleteDependent = function deleteDependent(dependentsMap, dependent) {
    dependentsMap.forEach(function (dependents) {
      dependents.delete(dependent);
    });
  };

  var setDependencies = function setDependencies(dependentsMap, atom, dependencies) {
    deleteDependent(dependentsMap, atom);
    dependencies.forEach(function (dependency) {
      addDependent(dependentsMap, dependency, atom);
    });
  };

  var listDependents = function listDependents(dependentsMap, atom, excludeSelf) {
    var dependents = new Set(dependentsMap.get(atom));

    if (excludeSelf) {
      dependents.delete(atom);
    }

    return dependents;
  };

  var initialState = new Map();

  var getAtomStateValue = function getAtomStateValue(atom, state, tmpState) {
    var atomState = tmpState && tmpState.get(atom) || state.get(atom);
    return atomState ? atomState.value : atom.init;
  };

  var readAtom = function readAtom(state, readingAtom, setState, dependentsMap) {
    var readAtomValue = function readAtomValue(prevState, atom) {
      var partialState = new Map();
      var atomState = prevState.get(atom);

      if (atomState) {
        return [atomState, partialState];
      }

      var error = undefined;
      var promise = undefined;
      var value = null;
      var dependencies = new Set();
      var isSync = true;

      try {
        var promiseOrValue = atom.read(function (a) {
          if (dependencies) {
            dependencies.add(a);
          } else {
            addDependent(dependentsMap, a, atom);
          }

          if (a !== atom) {
            var _readAtomValue = readAtomValue(prevState, a),
                _nextAtomState = _readAtomValue[0],
                nextPartialState = _readAtomValue[1];

            if (isSync) {
              appendMap(partialState, nextPartialState);
            } else {
              setState(function (prev) {
                return concatMap(prev, nextPartialState);
              });
            }

            if (_nextAtomState.error) {
              throw _nextAtomState.error;
            }

            if (_nextAtomState.promise) {
              throw _nextAtomState.promise;
            }

            return _nextAtomState.value;
          } // a === atom


          var aState = prevState.get(a);

          if (aState) {
            if (aState.promise) {
              throw aState.promise;
            }

            return aState.value;
          }

          return a.init; // this should not be undefined
        });

        if (promiseOrValue instanceof Promise) {
          promise = promiseOrValue.then(function (value) {
            setDependencies(dependentsMap, atom, dependencies);
            dependencies = null;
            setState(function (prev) {
              return new Map(prev).set(atom, {
                value: value
              });
            });
          }).catch(function (e) {
            setState(function (prev) {
              return new Map(prev).set(atom, {
                value: getAtomStateValue(atom, prev),
                error: e instanceof Error ? e : new Error(e)
              });
            });
          });
        } else {
          setDependencies(dependentsMap, atom, dependencies);
          dependencies = null;
          value = promiseOrValue;
        }
      } catch (errorOrPromise) {
        if (errorOrPromise instanceof Promise) {
          promise = errorOrPromise;
        } else if (errorOrPromise instanceof Error) {
          error = errorOrPromise;
        } else {
          error = new Error(errorOrPromise);
        }
      }

      var nextAtomState = {
        error: error,
        promise: promise,
        value: promise ? atom.init : value
      };
      partialState.set(atom, nextAtomState);
      isSync = false;
      return [nextAtomState, partialState];
    };

    return readAtomValue(state, readingAtom);
  };

  var addAtom = function addAtom(id, atom, partialState, setState, dependentsMap) {
    addDependent(dependentsMap, atom, id);

    if (partialState) {
      setState(function (prev) {
        return concatMap(prev, partialState);
      });
    }
  };

  var delAtom = function delAtom(id, setGcCount, dependentsMap) {
    deleteDependent(dependentsMap, id);
    setGcCount(function (c) {
      return c + 1;
    }); // trigger re-render for gc
  };

  var gcAtom = function gcAtom(state, setState, dependentsMap) {
    var gcAtomState = function gcAtomState(prevState) {
      var nextState = new Map(prevState);

      while (true) {
        var deleted = false;
        nextState.forEach(function (_atomState, atom) {
          var _dependentsMap$get;

          var isEmpty = ((_dependentsMap$get = dependentsMap.get(atom)) == null ? void 0 : _dependentsMap$get.size) === 0;

          if (isEmpty) {
            nextState.delete(atom);
            dependentsMap.delete(atom);
            deleted = true;
          }
        });

        if (!deleted) {
          break;
        }
      }

      return nextState;
    };

    var nextState = gcAtomState(state);

    if (state.size !== nextState.size) {
      setState(nextState);
    }
  };

  var writeAtom = function writeAtom(updatingAtom, update, setState, dependentsMap, addWriteThunk) {
    var pendingPromises = [];

    var updateDependentsState = function updateDependentsState(prevState, atom) {
      var partialState = new Map();
      listDependents(dependentsMap, atom, true).forEach(function (dependent) {
        if (typeof dependent === 'symbol') return;
        var dependencies = new Set();

        try {
          var v = dependent.read(function (a) {
            if (dependencies) {
              dependencies.add(a);
            } else {
              addDependent(dependentsMap, a, dependent);
            }

            var s = prevState.get(a);

            if (!s) {
              throw new Error('atom state not found. possibly a bug.');
            }

            if (s.error) {
              throw s.error;
            }

            return s.value;
          });

          if (v instanceof Promise) {
            var promise = v.then(function (vv) {
              setDependencies(dependentsMap, dependent, dependencies);
              dependencies = null;
              var nextAtomState = {
                value: vv
              };
              setState(function (prev) {
                var nextState = new Map(prev).set(dependent, nextAtomState);
                var nextPartialState = updateDependentsState(nextState, dependent);
                return appendMap(nextState, nextPartialState);
              });
            }).catch(function (e) {
              setState(function (prev) {
                return new Map(prev).set(dependent, {
                  value: getAtomStateValue(dependent, prev),
                  error: e instanceof Error ? e : new Error(e)
                });
              });
            });
            partialState.set(dependent, {
              value: getAtomStateValue(dependent, prevState),
              promise: promise
            });
          } else {
            setDependencies(dependentsMap, dependent, dependencies);
            dependencies = null;
            partialState.set(dependent, {
              value: v
            });
            appendMap(partialState, updateDependentsState(concatMap(prevState, partialState), dependent));
          }
        } catch (e) {
          partialState.set(dependent, {
            value: getAtomStateValue(dependent, prevState),
            error: e instanceof Error ? e : new Error(e)
          });
          appendMap(partialState, updateDependentsState(concatMap(prevState, partialState), dependent));
        }
      });
      return partialState;
    };

    var updateAtomState = function updateAtomState(prevState, atom, update) {
      var partialState = new Map();
      var isSync = true;

      try {
        var promise = atom.write(function (a) {
          if (process.env.NODE_ENV !== 'production') {
            var s = prevState.get(a);

            if (s && s.promise) {
              console.log('Reading pending atom state in write operation. Not sure how to deal with it. Returning obsolete vaule for', a);
            }
          }

          return getAtomStateValue(a, prevState, partialState);
        }, function (a, v) {
          if (a === atom) {
            var nextAtomState = {
              value: v
            };

            if (isSync) {
              partialState.set(a, nextAtomState);
              appendMap(partialState, updateDependentsState(concatMap(prevState, partialState), a));
            } else {
              setState(function (prev) {
                var nextState = new Map(prev).set(a, nextAtomState);
                var nextPartialState = updateDependentsState(nextState, a);
                return appendMap(nextState, nextPartialState);
              });
            }
          } else {
            if (isSync) {
              var nextPartialState = updateAtomState(prevState, a, v);
              appendMap(partialState, nextPartialState);
            } else {
              addWriteThunk(function (prev) {
                var nextPartialState = updateAtomState(prev, a, v);
                return concatMap(prev, nextPartialState);
              });
            }
          }
        }, update);

        if (promise instanceof Promise) {
          pendingPromises.push(promise); // XXX this is write pending (can be confused with read pending)

          var nextAtomState = {
            value: getAtomStateValue(atom, prevState, partialState),
            promise: promise.then(function () {
              addWriteThunk(function (prev) {
                return new Map(prev).set(atom, {
                  value: getAtomStateValue(atom, prev)
                });
              });
            })
          };
          partialState.set(atom, nextAtomState);
        }
      } catch (e) {
        if (pendingPromises.length) {
          pendingPromises.push(new Promise(function (_resolve, reject) {
            reject(e);
          }));
        } else {
          throw e;
        }
      }

      isSync = false;
      return partialState;
    };

    addWriteThunk(function (prevState) {
      var nextPartialState = updateAtomState(prevState, updatingAtom, update);
      return concatMap(prevState, nextPartialState);
    });

    if (pendingPromises.length) {
      return new Promise(function (resolve, reject) {
        var loop = function loop() {
          var len = pendingPromises.length;

          if (len === 0) {
            resolve();
          } else {
            Promise.all(pendingPromises).then(function () {
              pendingPromises.splice(0, len);
              loop();
            }).catch(reject);
          }
        };

        loop();
      });
    }
  };

  var runWriteThunk = function runWriteThunk(lastStateRef, setState, writeThunkQueue) {
    while (true) {
      if (lastStateRef.current === null) {
        return;
      }

      if (writeThunkQueue.length === 0) {
        return;
      }

      var thunk = writeThunkQueue.shift();
      var lastState = lastStateRef.current;
      var nextState = thunk(lastState);

      if (nextState !== lastState) {
        setState(nextState);
        return;
      }
    }
  };

  var ActionsContext = useContextSelector.createContext(warningObject);
  var StateContext = useContextSelector.createContext(warningObject);
  var Provider = function Provider(_ref) {
    var children = _ref.children;

    var _useState = react.useState(initialState),
        state = _useState[0],
        setStateOrig = _useState[1];

    var setState = function setState(setStateAction) {
      lastStateRef.current = null;
      setStateOrig(setStateAction);
    };

    var dependentsMapRef = react.useRef();

    if (!dependentsMapRef.current) {
      dependentsMapRef.current = new Map();
    }

    var _useState2 = react.useState(0),
        gcCount = _useState2[0],
        setGcCount = _useState2[1]; // to trigger gc


    react.useEffect(function () {
      gcAtom(state, setState, dependentsMapRef.current);
    }, [state, gcCount]);
    var lastStateRef = react.useRef(null);
    useIsoLayoutEffect(function () {
      lastStateRef.current = state;
    });
    var writeThunkQueueRef = react.useRef([]);
    react.useEffect(function () {
      runWriteThunk(lastStateRef, setState, writeThunkQueueRef.current);
    }, [state]);
    var actions = react.useMemo(function () {
      return {
        add: function add(id, atom, partialState) {
          return addAtom(id, atom, partialState, setState, dependentsMapRef.current);
        },
        del: function del(id) {
          return delAtom(id, setGcCount, dependentsMapRef.current);
        },
        read: function read(state, atom) {
          return readAtom(state, atom, setState, dependentsMapRef.current);
        },
        write: function write(atom, update) {
          return writeAtom(atom, update, setState, dependentsMapRef.current, function (thunk) {
            writeThunkQueueRef.current.push(thunk);
            runWriteThunk(lastStateRef, setState, writeThunkQueueRef.current);
          });
        }
      };
    }, []);
    return react.createElement(ActionsContext.Provider, {
      value: actions
    }, react.createElement(StateContext.Provider, {
      value: state
    }, children));
  };

  var keyCount = 0; // global key count for all atoms

  function atom(read, write) {
    var config = {
      key: ++keyCount
    };

    if (typeof read === 'function') {
      config.read = read;
    } else {
      config.init = read;

      config.read = function (get) {
        return get(config);
      };

      config.write = function (get, set, update) {
        set(config, typeof update === 'function' ? update(get(config)) : update);
      };
    }

    if (write) {
      config.write = write;
    }

    return config;
  }

  var isWritable = function isWritable(atom) {
    return !!atom.write;
  };

  function useAtom(atom) {
    var actions = useContextSelector.useContext(ActionsContext);
    var pendingListRef = react.useRef([]);
    var value = useContextSelector.useContextSelector(StateContext, react.useCallback(function (state) {
      var atomState = state.get(atom);

      if (!atomState) {
        var _actions$read = actions.read(state, atom),
            initialAtomState = _actions$read[0],
            pendingPartialState = _actions$read[1];

        atomState = initialAtomState;

        if (!atomState.error && !atomState.promise && pendingPartialState.size) {
          pendingListRef.current.unshift({
            v: initialAtomState.value,
            p: pendingPartialState
          });
        }
      }

      if (atomState.error) {
        throw atomState.error;
      }

      if (atomState.promise) {
        throw atomState.promise;
      }

      return atomState.value;
    }, [atom, actions]));
    var pendingPartialStateRef = react.useRef();
    useIsoLayoutEffect(function () {
      var pendingList = pendingListRef.current;
      var found = pendingList.find(function (_ref) {
        var v = _ref.v;
        return v === value;
      });

      if (found) {
        pendingPartialStateRef.current = found.p;
      }

      pendingList.splice(0, pendingList.length);
    });
    useIsoLayoutEffect(function () {
      var id = Symbol();
      actions.add(id, atom, pendingPartialStateRef.current);
      return function () {
        actions.del(id);
      };
    }, [actions, atom]);
    var setAtom = react.useCallback(function (update) {
      if (isWritable(atom)) {
        return actions.write(atom, update);
      } else {
        throw new Error('not writable atom');
      }
    }, [atom, actions]);
    react.useDebugValue(value);
    return [value, setAtom];
  }

  exports.Provider = Provider;
  exports.atom = atom;
  exports.useAtom = useAtom;

  return exports;

}({}, React, useContextSelector));
