import { useLayoutEffect, useState, useRef, useEffect, useMemo, createElement, useCallback, useDebugValue } from 'react';
import { createContext, useContext, useContextSelector } from 'use-context-selector';

const isSSR = typeof window === 'undefined' || /ServerSideRendering/.test(window.navigator && window.navigator.userAgent);
const useIsoLayoutEffect = isSSR ? fn => fn() : useLayoutEffect;

const appendMap = (dst, src) => {
  src.forEach((v, k) => {
    dst.set(k, v);
  });
  return dst;
}; // create new map from two maps


const concatMap = (src1, src2) => appendMap(new Map(src1), src2);

const warningObject = new Proxy({}, {
  get() {
    throw new Error('Please use <Provider>');
  },

  apply() {
    throw new Error('Please use <Provider>');
  }

});

const addDependent = (dependentsMap, atom, dependent) => {
  let dependents = dependentsMap.get(atom);

  if (!dependents) {
    dependents = new Set();
    dependentsMap.set(atom, dependents);
  }

  dependents.add(dependent);
};

const deleteDependent = (dependentsMap, dependent) => {
  dependentsMap.forEach(dependents => {
    dependents.delete(dependent);
  });
};

const setDependencies = (dependentsMap, atom, dependencies) => {
  deleteDependent(dependentsMap, atom);
  dependencies.forEach(dependency => {
    addDependent(dependentsMap, dependency, atom);
  });
};

const listDependents = (dependentsMap, atom, excludeSelf) => {
  const dependents = new Set(dependentsMap.get(atom));

  if (excludeSelf) {
    dependents.delete(atom);
  }

  return dependents;
};

const initialState = new Map();

const getAtomStateValue = (atom, state, tmpState) => {
  const atomState = tmpState && tmpState.get(atom) || state.get(atom);
  return atomState ? atomState.value : atom.init;
};

const readAtom = (state, readingAtom, setState, dependentsMap) => {
  const readAtomValue = (prevState, atom) => {
    const partialState = new Map();
    const atomState = prevState.get(atom);

    if (atomState) {
      return [atomState, partialState];
    }

    let error = undefined;
    let promise = undefined;
    let value = null;
    let dependencies = new Set();
    let isSync = true;

    try {
      const promiseOrValue = atom.read(a => {
        if (dependencies) {
          dependencies.add(a);
        } else {
          addDependent(dependentsMap, a, atom);
        }

        if (a !== atom) {
          const [nextAtomState, nextPartialState] = readAtomValue(prevState, a);

          if (isSync) {
            appendMap(partialState, nextPartialState);
          } else {
            setState(prev => concatMap(prev, nextPartialState));
          }

          if (nextAtomState.error) {
            throw nextAtomState.error;
          }

          if (nextAtomState.promise) {
            throw nextAtomState.promise;
          }

          return nextAtomState.value;
        } // a === atom


        const aState = prevState.get(a);

        if (aState) {
          if (aState.promise) {
            throw aState.promise;
          }

          return aState.value;
        }

        return a.init; // this should not be undefined
      });

      if (promiseOrValue instanceof Promise) {
        promise = promiseOrValue.then(value => {
          setDependencies(dependentsMap, atom, dependencies);
          dependencies = null;
          setState(prev => new Map(prev).set(atom, {
            value
          }));
        }).catch(e => {
          setState(prev => new Map(prev).set(atom, {
            value: getAtomStateValue(atom, prev),
            error: e instanceof Error ? e : new Error(e)
          }));
        });
      } else {
        setDependencies(dependentsMap, atom, dependencies);
        dependencies = null;
        value = promiseOrValue;
      }
    } catch (errorOrPromise) {
      if (errorOrPromise instanceof Promise) {
        promise = errorOrPromise;
      } else if (errorOrPromise instanceof Error) {
        error = errorOrPromise;
      } else {
        error = new Error(errorOrPromise);
      }
    }

    const nextAtomState = {
      error,
      promise,
      value: promise ? atom.init : value
    };
    partialState.set(atom, nextAtomState);
    isSync = false;
    return [nextAtomState, partialState];
  };

  return readAtomValue(state, readingAtom);
};

const addAtom = (id, atom, partialState, setState, dependentsMap) => {
  addDependent(dependentsMap, atom, id);

  if (partialState) {
    setState(prev => concatMap(prev, partialState));
  }
};

const delAtom = (id, setGcCount, dependentsMap) => {
  deleteDependent(dependentsMap, id);
  setGcCount(c => c + 1); // trigger re-render for gc
};

const gcAtom = (state, setState, dependentsMap) => {
  const gcAtomState = prevState => {
    const nextState = new Map(prevState);

    while (true) {
      let deleted = false;
      nextState.forEach((_atomState, atom) => {
        var _dependentsMap$get;

        const isEmpty = ((_dependentsMap$get = dependentsMap.get(atom)) == null ? void 0 : _dependentsMap$get.size) === 0;

        if (isEmpty) {
          nextState.delete(atom);
          dependentsMap.delete(atom);
          deleted = true;
        }
      });

      if (!deleted) {
        break;
      }
    }

    return nextState;
  };

  const nextState = gcAtomState(state);

  if (state.size !== nextState.size) {
    setState(nextState);
  }
};

const writeAtom = (updatingAtom, update, setState, dependentsMap, addWriteThunk) => {
  const pendingPromises = [];

  const updateDependentsState = (prevState, atom) => {
    const partialState = new Map();
    listDependents(dependentsMap, atom, true).forEach(dependent => {
      if (typeof dependent === 'symbol') return;
      let dependencies = new Set();

      try {
        const v = dependent.read(a => {
          if (dependencies) {
            dependencies.add(a);
          } else {
            addDependent(dependentsMap, a, dependent);
          }

          const s = prevState.get(a);

          if (!s) {
            throw new Error('atom state not found. possibly a bug.');
          }

          if (s.error) {
            throw s.error;
          }

          return s.value;
        });

        if (v instanceof Promise) {
          const promise = v.then(vv => {
            setDependencies(dependentsMap, dependent, dependencies);
            dependencies = null;
            const nextAtomState = {
              value: vv
            };
            setState(prev => {
              const nextState = new Map(prev).set(dependent, nextAtomState);
              const nextPartialState = updateDependentsState(nextState, dependent);
              return appendMap(nextState, nextPartialState);
            });
          }).catch(e => {
            setState(prev => new Map(prev).set(dependent, {
              value: getAtomStateValue(dependent, prev),
              error: e instanceof Error ? e : new Error(e)
            }));
          });
          partialState.set(dependent, {
            value: getAtomStateValue(dependent, prevState),
            promise
          });
        } else {
          setDependencies(dependentsMap, dependent, dependencies);
          dependencies = null;
          partialState.set(dependent, {
            value: v
          });
          appendMap(partialState, updateDependentsState(concatMap(prevState, partialState), dependent));
        }
      } catch (e) {
        partialState.set(dependent, {
          value: getAtomStateValue(dependent, prevState),
          error: e instanceof Error ? e : new Error(e)
        });
        appendMap(partialState, updateDependentsState(concatMap(prevState, partialState), dependent));
      }
    });
    return partialState;
  };

  const updateAtomState = (prevState, atom, update) => {
    const partialState = new Map();
    let isSync = true;

    try {
      const promise = atom.write(a => {
        if (process.env.NODE_ENV !== 'production') {
          const s = prevState.get(a);

          if (s && s.promise) {
            console.log('Reading pending atom state in write operation. Not sure how to deal with it. Returning obsolete vaule for', a);
          }
        }

        return getAtomStateValue(a, prevState, partialState);
      }, (a, v) => {
        if (a === atom) {
          const nextAtomState = {
            value: v
          };

          if (isSync) {
            partialState.set(a, nextAtomState);
            appendMap(partialState, updateDependentsState(concatMap(prevState, partialState), a));
          } else {
            setState(prev => {
              const nextState = new Map(prev).set(a, nextAtomState);
              const nextPartialState = updateDependentsState(nextState, a);
              return appendMap(nextState, nextPartialState);
            });
          }
        } else {
          if (isSync) {
            const nextPartialState = updateAtomState(prevState, a, v);
            appendMap(partialState, nextPartialState);
          } else {
            addWriteThunk(prev => {
              const nextPartialState = updateAtomState(prev, a, v);
              return concatMap(prev, nextPartialState);
            });
          }
        }
      }, update);

      if (promise instanceof Promise) {
        pendingPromises.push(promise); // XXX this is write pending (can be confused with read pending)

        const nextAtomState = {
          value: getAtomStateValue(atom, prevState, partialState),
          promise: promise.then(() => {
            addWriteThunk(prev => new Map(prev).set(atom, {
              value: getAtomStateValue(atom, prev)
            }));
          })
        };
        partialState.set(atom, nextAtomState);
      }
    } catch (e) {
      if (pendingPromises.length) {
        pendingPromises.push(new Promise((_resolve, reject) => {
          reject(e);
        }));
      } else {
        throw e;
      }
    }

    isSync = false;
    return partialState;
  };

  addWriteThunk(prevState => {
    const nextPartialState = updateAtomState(prevState, updatingAtom, update);
    return concatMap(prevState, nextPartialState);
  });

  if (pendingPromises.length) {
    return new Promise((resolve, reject) => {
      const loop = () => {
        const len = pendingPromises.length;

        if (len === 0) {
          resolve();
        } else {
          Promise.all(pendingPromises).then(() => {
            pendingPromises.splice(0, len);
            loop();
          }).catch(reject);
        }
      };

      loop();
    });
  }
};

const runWriteThunk = (lastStateRef, setState, writeThunkQueue) => {
  while (true) {
    if (lastStateRef.current === null) {
      return;
    }

    if (writeThunkQueue.length === 0) {
      return;
    }

    const thunk = writeThunkQueue.shift();
    const lastState = lastStateRef.current;
    const nextState = thunk(lastState);

    if (nextState !== lastState) {
      setState(nextState);
      return;
    }
  }
};

const ActionsContext = createContext(warningObject);
const StateContext = createContext(warningObject);
const Provider = ({
  children
}) => {
  const [state, setStateOrig] = useState(initialState);

  const setState = setStateAction => {
    lastStateRef.current = null;
    setStateOrig(setStateAction);
  };

  const dependentsMapRef = useRef();

  if (!dependentsMapRef.current) {
    dependentsMapRef.current = new Map();
  }

  const [gcCount, setGcCount] = useState(0); // to trigger gc

  useEffect(() => {
    gcAtom(state, setState, dependentsMapRef.current);
  }, [state, gcCount]);
  const lastStateRef = useRef(null);
  useIsoLayoutEffect(() => {
    lastStateRef.current = state;
  });
  const writeThunkQueueRef = useRef([]);
  useEffect(() => {
    runWriteThunk(lastStateRef, setState, writeThunkQueueRef.current);
  }, [state]);
  const actions = useMemo(() => ({
    add: (id, atom, partialState) => addAtom(id, atom, partialState, setState, dependentsMapRef.current),
    del: id => delAtom(id, setGcCount, dependentsMapRef.current),
    read: (state, atom) => readAtom(state, atom, setState, dependentsMapRef.current),
    write: (atom, update) => writeAtom(atom, update, setState, dependentsMapRef.current, thunk => {
      writeThunkQueueRef.current.push(thunk);
      runWriteThunk(lastStateRef, setState, writeThunkQueueRef.current);
    })
  }), []);
  return createElement(ActionsContext.Provider, {
    value: actions
  }, createElement(StateContext.Provider, {
    value: state
  }, children));
};

let keyCount = 0; // global key count for all atoms

function atom(read, write) {
  const config = {
    key: ++keyCount
  };

  if (typeof read === 'function') {
    config.read = read;
  } else {
    config.init = read;

    config.read = get => get(config);

    config.write = (get, set, update) => {
      set(config, typeof update === 'function' ? update(get(config)) : update);
    };
  }

  if (write) {
    config.write = write;
  }

  return config;
}

const isWritable = atom => !!atom.write;

function useAtom(atom) {
  const actions = useContext(ActionsContext);
  const pendingListRef = useRef([]);
  const value = useContextSelector(StateContext, useCallback(state => {
    let atomState = state.get(atom);

    if (!atomState) {
      const [initialAtomState, pendingPartialState] = actions.read(state, atom);
      atomState = initialAtomState;

      if (!atomState.error && !atomState.promise && pendingPartialState.size) {
        pendingListRef.current.unshift({
          v: initialAtomState.value,
          p: pendingPartialState
        });
      }
    }

    if (atomState.error) {
      throw atomState.error;
    }

    if (atomState.promise) {
      throw atomState.promise;
    }

    return atomState.value;
  }, [atom, actions]));
  const pendingPartialStateRef = useRef();
  useIsoLayoutEffect(() => {
    const pendingList = pendingListRef.current;
    const found = pendingList.find(({
      v
    }) => v === value);

    if (found) {
      pendingPartialStateRef.current = found.p;
    }

    pendingList.splice(0, pendingList.length);
  });
  useIsoLayoutEffect(() => {
    const id = Symbol();
    actions.add(id, atom, pendingPartialStateRef.current);
    return () => {
      actions.del(id);
    };
  }, [actions, atom]);
  const setAtom = useCallback(update => {
    if (isWritable(atom)) {
      return actions.write(atom, update);
    } else {
      throw new Error('not writable atom');
    }
  }, [atom, actions]);
  useDebugValue(value);
  return [value, setAtom];
}

export { Provider, atom, useAtom };
